import contextlib
import functools
import logging
import sys

from . import external_work, logmux

logger = logging.getLogger(__name__)


def redirected_job(decorate_out=None, decorate_err=None, redirect_logger=True):
    """Setup stdout/stderr redirection for the given coroutine.

    This is the same as wrapping the entire corouting in:

        async with ctx.setup_redirection(...):
            ...

    Use as a decorator:

        @redirected_job()
        async def my_job(ctx):
            ctx.logger.info('hello')
            await asyncio.sleep(1)
            ctx.logger.info(world')

        @redirected_job(decorate_out=lambda s: f'DECORATED: {s}')
        async def my_other_job(ctx):
            print('hello', file=ctx.stdout)
            await asyncio.sleep(1)
            print(again', file=ctx.stdout)
    """

    def wrap(coro):
        @functools.wraps(coro)
        async def wrapped_coro(ctx):
            async with ctx.setup_redirection(
                decorate_out=decorate_out,
                decorate_err=decorate_err,
                redirect_logger=redirect_logger,
            ):
                return await coro(ctx)

        return wrapped_coro

    return wrap


class Context(external_work.Context):
    """API for jobs with stdout/stderr redirected via a LogMux-aware Scheduler.

    This enables jobs to have their output multiplexed to a single (pair of)
    output stream(s) controlled by the below Scheduler.

    Redirection of the actual stdout/stderr file descriptors is automatically
    done for subprocesses, and for self.logger (unless redirect_logger is set
    to False). Other output (from thread workers or directly from .__call__()
    must be redirected to self.stdout/self.stderr manually.)
    """

    def __init__(self, *args):
        super().__init__(*args)
        self.stdout = None
        self.stderr = None

    @contextlib.asynccontextmanager
    async def setup_redirection(
        self, *, decorate_out=None, decorate_err=None, redirect_logger=True
    ):
        async with self._scheduler.outmux.new_stream(decorate_out) as outf:
            async with self._scheduler.errmux.new_stream(decorate_err) as errf:
                self.stdout = outf
                self.stderr = errf
                log_handler = None
                if redirect_logger:
                    log_handler = logging.StreamHandler(self.stderr)
                    self.logger.addHandler(log_handler)
                try:
                    yield
                finally:
                    if log_handler is not None:
                        self.logger.removeHandler(log_handler)
                    self.stdout = None
                    self.stderr = None

    async def run_in_subprocess(self, argv, **kwargs):
        """Pass redirected out/err as stdout/stderr to the subprocess.

        Only if stdout/stderr is not already customized by the caller.
        """
        if kwargs.get('stdout') is None:
            kwargs['stdout'] = self.stdout
        if kwargs.get('stderr') is None:
            kwargs['stderr'] = self.stderr
        return await super().run_in_subprocess(argv, **kwargs)


class Scheduler(external_work.Scheduler):
    """Run jobs with their output multiplexed through LogMux.

    Extend Scheduler with two LogMux instances - one for stdout and one for
    stderr. Can be used by job coroutines redirect their stdout/stderr through
    to a single pair of output streams controlled by the caller.
    """

    def __init__(
        self, *, outmux=None, errmux=None, context_class=Context, **kwargs
    ):
        self.outmux = logmux.LogMux(sys.stdout) if outmux is None else outmux
        self.errmux = logmux.LogMux(sys.stderr) if errmux is None else errmux

        assert issubclass(context_class, Context)
        super().__init__(context_class=context_class, **kwargs)

    async def _run_tasks(self, *args, **kwargs):
        logger.debug('Starting LogMux instances…')
        async with self.outmux:
            async with self.errmux:
                await super()._run_tasks(*args, **kwargs)



"""Verify expected events from the Scheduler.

Allow test jobs to declare a sequence of expected events, and verify that
these events are indeed emitted by the Scheduler.
"""
import time

# Wildcard used for don't-care values in expected event fields
Whatever = object()


def verify_one(expect, actual):
    """Verify one event against its expected counterpart."""
    expect_keys = set(expect.keys()) - {'MAY_CANCEL'}
    assert expect_keys == set(actual.keys())
    keys = sorted(expect_keys)
    for e, a in zip((expect[k] for k in keys), (actual[k] for k in keys)):
        if e is not Whatever:
            assert e == a
    return True


class ExpectedJobEvents:
    """Keep track of expected events per job."""

    def __init__(self, job_name):
        self.name = job_name
        self.expected = []
        self.verifying = False

    def add(self, event, *, may_cancel=False, **kwargs):
        assert not self.verifying
        d = {'event': event, 'job': self.name, 'timestamp': Whatever}
        d.update(kwargs)
        if may_cancel:
            d['MAY_CANCEL'] = True
        self.expected.append(d)

    def end(self):
        """Call this when the job ends, before we start verifying events."""
        assert not self.verifying
        if self.expected[-1]['event'] != 'finish':  # we were cancelled
            self.add('finish', fate='cancelled')
        self.verifying = True

    def verify_next(self, actual):
        assert self.verifying
        if actual['event'] == 'finish' and actual['fate'] == 'cancelled':
            while self.expected[0].get('MAY_CANCEL', False):
                self.expected.pop(0)
        return verify_one(self.expected.pop(0), actual)

    def __len__(self):
        return len(self.expected)


class EventVerifier:
    """Keep track of emitted events."""

    def __init__(self):
        self._start = time.time()
        self._end = None
        self.events = []

    def __call__(self, event):
        assert self._end is None
        self.events.append(event)

    def end(self):
        assert self._end is None
        self._end = time.time()

    def _verify_timestamps(self):
        """Timestamps are in sorted order and all between ._start and ._end."""
        assert self._end is not None
        timestamps = [e['timestamp'] for e in self.events]
        assert timestamps == sorted(timestamps)
        assert timestamps[0] >= self._start
        assert timestamps[-1] <= self._end

    def _verify_initial_adds(self, initial_xjobs):
        """Jobs added before Scheduler.run() emit 'add' events in order.

        Verify that the 'n' first events are 'add' events for the first 'n'
        jobs, in order.
        """
        for i in range(len(initial_xjobs)):
            actual = self.events.pop(0)
            assert actual['event'] == 'add'
            assert initial_xjobs[i].verify_next(actual)

    def _verify_overall_start_and_finish(self, num_initial, num_total):
        """Overall execution is bookended by start and finish events."""
        first, last = self.events.pop(0), self.events.pop()
        assert verify_one(
            {
                'event': 'start',
                'num_jobs': num_initial,
                'keep_going': Whatever,
                'timestamp': Whatever,
            },
            first,
        )
        assert verify_one(
            {'event': 'finish', 'num_tasks': num_total, 'timestamp': Whatever},
            last,
        )

    def _verify_initial_starts(self, initial_xjobs):
        """Initial jobs are started when Scheduler starts running."""
        for i in range(len(initial_xjobs)):
            actual = self.events.pop(0)
            assert actual['event'] == 'start'
            assert initial_xjobs[i].verify_next(actual)

    def _verify_await_tasks(self, initial_job_names):
        """Task running is bookended by "await/awaited tasks" events."""
        first, last = self.events.pop(0), self.events.pop()
        assert verify_one(
            {
                'event': 'await tasks',
                'jobs': initial_job_names,
                'timestamp': Whatever,
            },
            first,
        )
        assert verify_one(
            {'event': 'awaited tasks', 'timestamp': Whatever}, last,
        )

    def verify_all(self, initial_xevents, spawned_xevents):
        """Verify all events emitted by a scheduler instance."""
        # Transition into verifying state
        if self._end is None:
            self.end()
        for xjob in initial_xevents + spawned_xevents:
            xjob.end()

        total_jobs = len(initial_xevents) + len(spawned_xevents)

        self._verify_timestamps()
        self._verify_initial_adds(initial_xevents)
        self._verify_overall_start_and_finish(len(initial_xevents), total_jobs)

        if initial_xevents or spawned_xevents:
            self._verify_initial_starts(initial_xevents)
            self._verify_await_tasks([xjob.name for xjob in initial_xevents])

            # Remaining events belong to individual jobs; this includes
            # expected events from jobs in spawned_xevents.
            xjobs = {x.name: x for x in initial_xevents + spawned_xevents}
            assert len(xjobs) == total_jobs

            # Verify each event by pairing it w/that job's next expected event
            # (we cannot deterministically sequence events between jobs)
            while self.events:
                actual = self.events.pop(0)
                xjobs[actual['job']].verify_next(actual)

            # No more expected events
            assert all(len(xjob) == 0 for xjob in xjobs.values())

        assert not self.events  # no more actual events
        return True



import pytest
from subprocess import CalledProcessError

from asyncjobs import external_work

from conftest import (
    Cancelled,
    TExternalWorkJob,
    verified_events,
    verify_tasks,
)

pytestmark = pytest.mark.asyncio

TJob = TExternalWorkJob


@pytest.fixture
def Scheduler(scheduler_with_workers):
    return scheduler_with_workers(external_work.Scheduler)


@pytest.fixture
def run(Scheduler):
    async def _run(todo):
        scheduler = Scheduler()
        with verified_events(scheduler, todo):
            for job in todo:
                scheduler.add_job(job.name, job, getattr(job, 'deps', None))
            return await scheduler.run()

    return _run


# simple scenarios with threads/subprocesses


async def test_one_ok_job_in_thread(run):
    todo = [TJob('foo', thread=lambda ctx: 'foo worked')]
    done = await run(todo)
    assert verify_tasks(done, {'foo': 'foo worked'})


async def test_one_ok_job_in_subproc(run, tmp_path):
    path = tmp_path / 'foo'
    todo = [TJob('foo', subproc=['touch', str(path)])]
    done = await run(todo)
    assert verify_tasks(done, {'foo': 0})
    assert path.is_file()


async def test_one_failed_between_two_ok_jobs_in_threads_cancels_last(run):
    def raiseUGH(ctx):
        raise ValueError('UGH')

    todo = [
        TJob('foo', before={'bar'}, thread=lambda ctx: 'foo worked'),
        TJob('bar', {'foo'}, before={'baz'}, thread=raiseUGH),
        TJob('baz', {'bar'}, thread=lambda ctx: 'baz worked'),
    ]
    done = await run(todo)
    assert verify_tasks(
        done, {'foo': 'foo worked', 'bar': ValueError('UGH'), 'baz': Cancelled}
    )


async def test_one_failed_between_two_in_subprocs_cancels_last(run, tmp_path):
    foo_path = tmp_path / 'foo'
    baz_path = tmp_path / 'baz'
    todo = [
        TJob('foo', before={'bar'}, subproc=['touch', str(foo_path)]),
        TJob('bar', {'foo'}, before={'baz'}, subproc=['false']),
        TJob('baz', {'bar'}, subproc=['touch', str(baz_path)]),
    ]
    done = await run(todo)
    assert verify_tasks(
        done,
        {'foo': 0, 'bar': CalledProcessError(1, ['false']), 'baz': Cancelled},
    )
    assert foo_path.is_file()
    assert not baz_path.exists()


# add_thread_job() helper


async def test_one_threaded_job_before_another(Scheduler):
    def first_job():
        return 'return value from first job'

    def second_job():
        return 'return value from second job'

    scheduler = Scheduler()
    scheduler.add_thread_job('first_job', first_job)
    scheduler.add_thread_job('second_job', second_job, deps={'first_job'})
    done = await scheduler.run()

    assert verify_tasks(
        done,
        {
            'first_job': 'return value from first job',
            'second_job': 'return value from second job',
        },
    )


async def test_one_threaded_failure_cancels_the_next(Scheduler):
    def first_job():
        raise ValueError('FAIL')

    def second_job():
        return 'return value from second job'

    scheduler = Scheduler()
    scheduler.add_thread_job('first_job', first_job)
    scheduler.add_thread_job('second_job', second_job, deps={'first_job'})
    done = await scheduler.run()

    assert verify_tasks(
        done, {'first_job': ValueError('FAIL'), 'second_job': Cancelled},
    )


# add_subprocess_job() helper


async def test_one_subproc_job_before_another(Scheduler):
    scheduler = Scheduler()
    scheduler.add_subprocess_job('first_job', ['true'])
    scheduler.add_subprocess_job('second_job', ['false'], deps={'first_job'})
    done = await scheduler.run()

    assert verify_tasks(
        done, {'first_job': 0, 'second_job': CalledProcessError(1, ['false'])},
    )


async def test_one_subprocess_failure_with_check_cancels_the_next(Scheduler):
    scheduler = Scheduler()
    scheduler.add_subprocess_job('first_job', ['false'], check=True)
    scheduler.add_subprocess_job('second_job', ['true'], deps={'first_job'})
    done = await scheduler.run()

    assert verify_tasks(
        done,
        {
            'first_job': CalledProcessError(1, ['false']),
            'second_job': Cancelled,
        },
    )


async def test_one_subprocess_failure_wo_check_proceeds_to_next(Scheduler):
    scheduler = Scheduler()
    scheduler.add_subprocess_job('first_job', ['false'], check=False)
    scheduler.add_subprocess_job('second_job', ['true'], deps={'first_job'})
    done = await scheduler.run()

    assert verify_tasks(done, {'first_job': 1, 'second_job': 0})


async def test_one_subprocess_failure_by_default_cancels_the_next(Scheduler):
    scheduler = Scheduler()
    scheduler.add_subprocess_job('first_job', ['false'])
    scheduler.add_subprocess_job('second_job', ['true'], deps={'first_job'})
    done = await scheduler.run()

    assert verify_tasks(
        done,
        {
            'first_job': CalledProcessError(1, ['false']),
            'second_job': Cancelled,
        },
    )



import asyncio
import concurrent.futures
import contextlib
import logging
import subprocess

from . import basic

logger = logging.getLogger(__name__)


class Context(basic.Context):
    """Extend Context with helpers for doing work in threads + processes."""

    @contextlib.asynccontextmanager
    async def reserve_worker(self):
        """Acquire a worker context where this job can run its own work.

        This is the mechanism that ensures we do not schedule more concurrent
        work than allowed by Scheduler.workers. Anybody that wants to spin off
        another thread or process to perform some work should use this context
        manager to await a free "slot".
        """
        self.logger.debug('-> acquiring worker semaphore…')
        self.event('await worker slot')
        async with self._scheduler.worker_semaphore:
            self.event('awaited worker slot')
            self.logger.debug('-- acquired worker semaphore')
            try:
                yield
            finally:
                self.logger.debug('<- releasing worker semaphore')

    async def call_in_thread(self, func, *args):
        """Call func(*args) in a worker thread and await its result."""
        async with self.reserve_worker():
            try:
                self.logger.debug(f'-> starting {func} in worker thread…')
                self.event('await worker thread', func=str(func))
                future = self._scheduler._start_in_thread(func, *args)
                future.add_done_callback(
                    lambda fut: self.event(
                        'awaited worker thread',
                        fate=self._scheduler._fate(fut),
                    )
                )
                self.logger.debug('-- awaiting worker thread…')
                result = await future
                self.logger.debug(f'<- {result!r} from worker')
                return result
            except Exception as e:
                self.logger.warning(f'<- Exception {e} from worker!')
                raise

    async def run_in_subprocess(
        self, argv, stdin=None, stdout=None, stderr=None, check=True
    ):
        """Run a command line in a subprocess and await its exit code."""
        returncode = None
        async with self.reserve_worker():
            self.logger.debug(f'-> starting {argv} in subprocess…')
            self.event('await worker proc', argv=argv)
            proc = await asyncio.create_subprocess_exec(
                *argv, stdin=stdin, stdout=stdout, stderr=stderr
            )
            try:
                self.logger.debug('-- awaiting subprocess…')
                returncode = await proc.wait()
            except asyncio.CancelledError:
                self.logger.error(f'Cancelled! Terminating {proc}!…')
                proc.terminate()
                try:
                    returncode = await proc.wait()
                    self.logger.debug(f'Cancelled! {proc} terminated.')
                except asyncio.CancelledError:
                    self.logger.error(f'Cancelled again! Killing {proc}!…')
                    proc.kill()
                    returncode = await proc.wait()
                    self.logger.debug(f'Cancelled! {proc} killed.')
                raise
            finally:
                self.event('awaited worker proc', returncode=returncode)
        if check and returncode != 0:
            raise subprocess.CalledProcessError(returncode, argv)
        return returncode


class Scheduler(basic.Scheduler):
    """Manage jobs whose work is done in threads or subprocesses.

    Extend Scheduler with methods that allow job coroutines to perform work in
    other threads or subprocesses, while keeping the number of _concurrent_
    threads/processes within the given limit.
    """

    def __init__(self, *, workers=1, context_class=Context, **kwargs):
        assert workers > 0
        self.workers = workers
        self.worker_semaphore = None
        self.worker_threads = None

        assert issubclass(context_class, Context)
        super().__init__(context_class=context_class, **kwargs)

    def add_thread_job(self, name, func, *args, deps=None):
        async def coro(ctx):
            return await ctx.call_in_thread(func, *args)

        return self.add_job(name, coro, deps)

    def add_subprocess_job(self, name, argv, *, deps=None, **kwargs):
        async def coro(ctx):
            return await ctx.run_in_subprocess(argv, **kwargs)

        return self.add_job(name, coro, deps)

    def _start_in_thread(self, func, *args):
        if self.worker_threads is None:
            self.worker_threads = concurrent.futures.ThreadPoolExecutor(
                max_workers=self.workers, thread_name_prefix='Worker Thread',
            )
        return asyncio.get_running_loop().run_in_executor(
            self.worker_threads, func, *args
        )

    async def _run_tasks(self, *args, **kwargs):
        self.worker_semaphore = asyncio.BoundedSemaphore(self.workers)

        try:
            await super()._run_tasks(*args, **kwargs)
            if self.worker_threads is not None:
                logger.debug('Shutting down worker threads…')
                self.worker_threads.shutdown()  # wait=timeout is None)
                self.worker_threads = None
                logger.debug('Shut down worker threads')
        finally:
            if self.worker_threads is not None:
                logger.error('Cancelling without awaiting workers…')
                self.worker_threads.shutdown(wait=False)
                self.worker_threads = None



import asyncio
import contextlib
import functools
from pathlib import Path
import pytest
import random
import time

from asyncjobs import logmuxed_work, signal_handling

from conftest import (
    abort_in,
    assert_elapsed_time_within,
    TExternalWorkJob,
    verified_events,
)

pytestmark = pytest.mark.asyncio

sh_helper = Path(__file__).parent / 'test_5_logmuxed_work_helper.sh'


def shuffled_prints(out_f, err_f, out_str, err_str):
    """Print out_str -> out_f and err_str -> err_f in shuffled order.

    Return a stream of callables that together prints out_str -> out_f and
    err_str -> err_f such that the each char is printed in order, but the
    out_f and err_f prints are interleaved in a random order. E.g. this call:

        shuffled_prints(sys.stdout, sys.stderr, 'foo', 'bar')

    might result in the following (un-called) prints being yielded:

        print('b', file=sys.stderr, end='')
        print('f', file=sys.stdout, end='')
        print('o', file=sys.stdout, end='')
        print('a', file=sys.stderr, end='')
        print('o', file=sys.stdout, end='')
        print('r', file=sys.stderr, end='')

    This helper is an attempt to provoke any issues we might have regarding
    rescheduling/ordering of tasks and internal handling of file descriptors.
    """

    def prints(f, s):
        """Yield callables that print each char of s to f, in order."""
        for c in s:
            yield functools.partial(print, c, file=f, end='')

    def random_interleave(*iterators):
        """Yield from each iterator in random order until all are exhausted."""
        iterators = list(iterators)
        while iterators:
            i = random.randrange(len(iterators))
            try:
                yield next(iterators[i])
            except StopIteration:  # iterators[i] is exhausted
                del iterators[i]

    out_prints = prints(out_f, out_str)
    err_prints = prints(err_f, err_str)
    yield from random_interleave(out_prints, err_prints)


def print_out_err(out, err, *, sync):
    # out/err are strings or lists of strings. If strings, the chars in the
    # string will be printed one-by-one, if lists of strings, each string
    # will be printed with a trailing newline.
    if isinstance(out, list):  # list of strings
        out = [line + '\n' for line in out]
    if isinstance(err, list):  # list of strings
        err = [line + '\n' for line in err]

    async def print_out_err_async(ctx):
        for print_one in shuffled_prints(ctx.stdout, ctx.stderr, out, err):
            print_one()
            await asyncio.sleep(0)

    def print_out_err_sync(ctx):
        for print_one in shuffled_prints(ctx.stdout, ctx.stderr, out, err):
            print_one()
            time.sleep(0.001)

    return print_out_err_sync if sync else print_out_err_async


def decorators(job_name):
    return (
        lambda msg: f'{job_name}/out: {msg.rstrip()}\n',
        lambda msg: f'{job_name}/ERR: {msg.rstrip()}\n',
    )


class TJob(TExternalWorkJob):
    def __init__(
        self,
        *args,
        out='',
        err='',
        in_thread=False,
        decorate=False,
        redirect_logger=False,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        if in_thread:
            self.thread = print_out_err(out, err, sync=True)
        elif 'subproc' not in kwargs:
            self.do_work = print_out_err(out, err, sync=False)
        self.redirect_logger = redirect_logger
        self.decorate_out, self.decorate_err = None, None
        if decorate:
            self.decorate_out, self.decorate_err = decorators(self.name)

    async def __call__(self, ctx):
        async with ctx.setup_redirection(
            decorate_out=self.decorate_out,
            decorate_err=self.decorate_err,
            redirect_logger=self.redirect_logger,
        ):
            return await super().__call__(ctx)


@pytest.fixture
def Scheduler(scheduler_with_workers):
    return scheduler_with_workers(
        signal_handling.Scheduler, logmuxed_work.Scheduler
    )


@pytest.fixture
def run(Scheduler):
    async def _run(todo, abort_after=None, check_events=True):
        scheduler = Scheduler()
        if check_events:
            cm = verified_events(scheduler, todo)
        else:
            cm = contextlib.nullcontext()
        with cm:
            for job in todo:
                scheduler.add_job(job.name, job, getattr(job, 'deps', None))
            with abort_in(abort_after):
                return await scheduler.run()

    return _run


# no output


async def test_no_output_from_no_jobs(run, verify_output):
    await run([])
    assert verify_output([], [])


async def test_no_output_from_two_jobs(run, verify_output):
    await run([TJob('foo'), TJob('bar')])
    assert verify_output([[], []], [[], []])


# undecorated output


async def test_default_output_is_undecorated(run, verify_output):
    jobs = ['foo', 'bar']
    out = "This is {name}'s stdout"
    err = "This is {name}'s stderr"

    todo = [
        # Pass out/err as lists of strings to print entire strings at once
        TJob(name, out=[out.format(name=name)], err=[err.format(name=name)])
        for name in jobs
    ]
    await run(todo)
    assert verify_output(
        [[out.format(name=name)] for name in jobs],
        [[err.format(name=name)] for name in jobs],
    )


# decorated output


async def test_decorated_output_from_two_jobs(run, verify_output):
    jobs = ['foo', 'bar']
    out = "This is {name}'s stdout"
    err = "This is {name}'s stderr"

    todo = [
        TJob(
            name,
            out=out.format(name=name),
            err=err.format(name=name),
            decorate=True,
        )
        for name in jobs
    ]
    await run(todo)

    outprefix = '{name}/out: '
    errprefix = '{name}/ERR: '
    assert verify_output(
        [[(outprefix + out).format(name=name)] for name in jobs],
        [[(errprefix + err).format(name=name)] for name in jobs],
    )


async def test_decorated_output_from_spawned_jobs(run, verify_output):
    foo = TJob(
        'foo',
        out="This is foo's stdout",
        err="This is foo's stderr",
        decorate=True,
        spawn=[
            TJob(
                'bar',
                out="This is bar's stdout",
                err="This is bar's stderr",
                decorate=True,
                spawn=[
                    TJob(
                        'baz',
                        out="This is baz's stdout",
                        err="This is baz's stderr",
                        decorate=True,
                    )
                ],
            )
        ],
    )
    await run([foo])
    assert verify_output(
        [
            ["foo/out: This is foo's stdout"],
            ["bar/out: This is bar's stdout"],
            ["baz/out: This is baz's stdout"],
        ],
        [
            ["foo/ERR: This is foo's stderr"],
            ["bar/ERR: This is bar's stderr"],
            ["baz/ERR: This is baz's stderr"],
        ],
    )


async def test_decorated_output_from_thread_worker(run, verify_output):
    jobs = ['foo', 'bar']
    out = "This is {name}'s stdout"
    err = "This is {name}'s stderr"

    todo = [
        TJob(
            name,
            out=out.format(name=name),
            err=err.format(name=name),
            decorate=True,
            in_thread=True,
        )
        for name in jobs
    ]
    await run(todo)

    outprefix = '{name}/out: '
    errprefix = '{name}/ERR: '
    assert verify_output(
        [[(outprefix + out).format(name=name)] for name in jobs],
        [[(errprefix + err).format(name=name)] for name in jobs],
    )


async def test_decorated_output_from_subprocess_worker(run, verify_output):
    assert sh_helper.is_file()
    jobs = ['foo', 'bar']
    out = "This is {name}'s stdout"
    err = "This is {name}'s stderr"

    todo = [
        TJob(
            name,
            decorate=True,
            subproc=[
                str(sh_helper),
                out.format(name=name),
                err.format(name=name),
            ],
        )
        for name in jobs
    ]
    await run(todo)

    outprefix = '{name}/out: '
    errprefix = '{name}/ERR: '
    assert verify_output(
        [[(outprefix + out).format(name=name)] for name in jobs],
        [[(errprefix + err).format(name=name)] for name in jobs],
    )


async def test_decorated_output_from_aborted_processes(
    num_workers, run, verify_output
):
    assert sh_helper.is_file()
    jobs = ['foo', 'bar', 'baz']
    out = "This is {name}'s stdout"
    err = "This is {name}'s stderr"

    todo = [
        TJob(
            name,
            decorate=True,
            subproc=[
                str(sh_helper),
                out.format(name=name),
                err.format(name=name),
                '1',  # sleep 1 sec between stdout print and stderr print
            ],
        )
        for name in jobs
    ]
    with assert_elapsed_time_within(0.75):
        await run(todo, abort_after=0.5)

    # We have 3 jobs, but can only run as many concurrently as there are
    # workers available. The rest will be cancelled before they start their
    # subprocess
    outprefix = '{name}/out: '
    expect_out = [[(outprefix + out).format(name=name)] for name in jobs]
    assert verify_output(expect_out[:num_workers], [])


# redirected_job() decorator


async def test_redirected_job_no_decoration(run, verify_output):
    @logmuxed_work.redirected_job()
    async def job(ctx):
        print('Printing to stdout', file=ctx.stdout)
        print('Printing to stderr', file=ctx.stderr)
        ctx.logger.info('Logging to stderr')

    job.name = 'foo'
    await run([job], check_events=False)
    assert verify_output(
        [['Printing to stdout']],
        [['Printing to stderr', 'Logging to stderr']],
    )


async def test_redirected_job_decorate_without_logger(run, verify_output):
    dec_out, dec_err = decorators('foo')

    @logmuxed_work.redirected_job(dec_out, dec_err, False)
    async def job(ctx):
        print('Printing to stdout', file=ctx.stdout)
        print('Printing to stderr', file=ctx.stderr)
        ctx.logger.info('Logging to stderr')

    job.name = 'foo'
    await run([job], check_events=False)
    assert verify_output(
        [['foo/out: Printing to stdout']], [['foo/ERR: Printing to stderr']],
    )


async def test_redirected_job_decorate_err_with_logger(run, verify_output):
    _, dec_err = decorators('foo')

    @logmuxed_work.redirected_job(decorate_err=dec_err)
    async def job(ctx):
        print('Printing to stdout', file=ctx.stdout)
        print('Printing to stderr', file=ctx.stderr)
        ctx.logger.info('Logging to stderr')

    job.name = 'foo'
    await run([job], check_events=False)
    assert verify_output(
        [['Printing to stdout']],
        [['foo/ERR: Printing to stderr', 'foo/ERR: Logging to stderr']],
    )


# stress-testing the logmux framework


async def test_decorated_output_from_100_jobs(run, verify_output):
    out = "This is {i}'s stdout"
    err = "This is {i}'s stderr"

    todo = [
        TJob(
            f'job #{i}',
            out=out.format(i=i),
            err=err.format(i=i),
            decorate=True,
        )
        for i in range(100)
    ]
    await run(todo)

    outprefix = 'job #{i}/out: '
    errprefix = 'job #{i}/ERR: '
    assert verify_output(
        [[(outprefix + out).format(i=i)] for i in range(100)],
        [[(errprefix + err).format(i=i)] for i in range(100)],
    )


async def test_decorated_output_from_100_spawned_jobs(run, verify_output):
    out = "This is {i}'s stdout"
    err = "This is {i}'s stderr"

    todo = TJob(
        'foo',
        spawn=[
            TJob(
                f'job #{i}',
                out=out.format(i=i),
                err=err.format(i=i),
                decorate=True,
            )
            for i in range(100)
        ],
    )
    await run([todo])

    outprefix = 'job #{i}/out: '
    errprefix = 'job #{i}/ERR: '
    assert verify_output(
        [[(outprefix + out).format(i=i)] for i in range(100)],
        [[(errprefix + err).format(i=i)] for i in range(100)],
    )


async def test_decorated_output_from_100_thread_workers(run, verify_output):
    out = "This is {i}'s stdout"
    err = "This is {i}'s stderr"

    todo = [
        TJob(
            f'job #{i}',
            out=out.format(i=i),
            err=err.format(i=i),
            decorate=True,
            in_thread=True,
        )
        for i in range(100)
    ]
    await run(todo)

    outprefix = 'job #{i}/out: '
    errprefix = 'job #{i}/ERR: '
    assert verify_output(
        [[(outprefix + out).format(i=i)] for i in range(100)],
        [[(errprefix + err).format(i=i)] for i in range(100)],
    )


async def test_decorated_output_from_100_subprocesses(run, verify_output):
    assert sh_helper.is_file()
    out = "This is {i}'s stdout"
    err = "This is {i}'s stderr"

    todo = [
        TJob(
            f'job #{i}',
            decorate=True,
            subproc=[str(sh_helper), out.format(i=i), err.format(i=i)],
        )
        for i in range(100)
    ]
    await run(todo)

    outprefix = 'job #{i}/out: '
    errprefix = 'job #{i}/ERR: '
    assert verify_output(
        [[(outprefix + out).format(i=i)] for i in range(100)],
        [[(errprefix + err).format(i=i)] for i in range(100)],
    )



#!/usr/bin/env python3
# Fetch a random wikipedia article and follow links from it
import asyncio
from bs4 import BeautifulSoup
from functools import partial
import requests
import sys

from asyncjobs import Scheduler

num_articles = 10
num_workers = 4


class Article:
    base_url = 'https://en.wikipedia.org'
    random_url = base_url + '/wiki/Special:Random'

    def __init__(self, url):
        resp = requests.get(url=url, allow_redirects=True)
        self.html = BeautifulSoup(resp.content, features='html.parser')

    def title(self):
        ret = self.html.title.string
        if ret.endswith('- Wikipedia'):
            ret = ret.rsplit('-', 1)[0].strip()
        return ret

    def hrefs(self):
        for p in self.html.find_all('p'):
            for a in p.find_all('a'):
                if a.get('href', '').startswith('/wiki/'):
                    yield self.base_url + a['href']


async def fetch(url, level, ctx):
    def fetch_in_thread():
        indent = '  ' * level
        print(f'{indent}  fetching {url}...')
        a = Article(url)
        title = a.title()
        hrefs = list(a.hrefs())
        print(f'{indent}* [{title}] links to {len(hrefs)} articles')
        return title, hrefs

    title, hrefs = await ctx.call_in_thread(fetch_in_thread)
    scheduler = ctx._scheduler
    for href in hrefs:
        if len(scheduler.jobs) < num_articles and href not in scheduler:
            ctx.add_job(href, partial(fetch, href, level + 1))


events = []
scheduler = Scheduler(workers=num_workers, event_handler=events.append)
scheduler.add_job(Article.random_url, partial(fetch, Article.random_url, 1))
asyncio.run(scheduler.run())
if 'plot' in sys.argv:
    from asyncjobs.plot_schedule import plot_schedule

    plot_schedule(
        f'Fetch {num_articles} Wikipedia articles with {num_workers} workers',
        events,
    ).show()



#!/usr/bin/env python3
# Run three simple jobs in sequence
import asyncio
from asyncjobs import Scheduler
import time


def sleep():  # Run in a worker thread by job #2 below
    print(f'{time.ctime()}: Sleep for a second')
    time.sleep(1)
    print(f'{time.ctime()}: Finished sleep')


s = Scheduler()

# Job #1 prints uptime
s.add_subprocess_job('#1', ['uptime'])

# Job #2 waits for #1 and then sleeps in a thread
s.add_thread_job('#2', sleep, deps={'#1'})

# Job #3 waits for #2 and then prints uptime (again)
s.add_subprocess_job('#3', ['uptime'], deps={'#2'})

asyncio.run(s.run())
